#include "tio.h"
#include "utils.h"
#include "tzipper.h"


bool CTioPlainText::load (const QString &fname)
{
  QFile file (fname);

  if (! file.open (QFile::ReadOnly | QFile::Text))
     {
      error_string = file.errorString();
      return false;
     }

  QTextStream in (&file);
  in.setCodec (charset.toUtf8());
  data = in.readAll();

  qDebug() << "tio load::CTioPlainText";

  return true;
}


bool CTioPlainText::save (const QString &fname)
{
  QFile file (fname);
  if (! file.open (QFile::WriteOnly | QFile::Text))
     {
      error_string = file.errorString();
      return false;
     }

  QTextStream out (&file);
  out.setCodec (charset.toUtf8());
  out << data;

  qDebug() << "tio save::CTioPlainText";

  return true;
}


CTioHandler::CTioHandler (QObject *parent): QObject (parent)
{
  default_handler = new CTioPlainText;
  list.append (default_handler);
  list.append (new CTioGzip);
  list.append (new CTioXMLBased);
  list.append (new CTioABW);
  list.append (new CTioRTF);
  list.append (new CTioSLA);

}


CTioHandler::~CTioHandler()
{
  while (! list.isEmpty())
       delete list.takeFirst();
}


CTio* CTioHandler::get_for_fname (const QString &fname)
{
  CTio *instance;
  QString ext = file_get_ext (fname).toLower();

  for (int i = 0; i < list.size(); i++)
      {
       instance = list.at (i);
       if (instance->extensions.indexOf (ext) != -1)
          return instance;
       }

  return default_handler;
}


CTioPlainText::CTioPlainText()
{
  ronly = false;
}


CTioGzip::CTioGzip()
{
  ronly = true;
  extensions.append ("gz");
}


bool CTioGzip::load (const QString &fname)
{
  QByteArray a = gzip_deflateFile (fname);
  data = a.data();

  qDebug() << "tio load::CTioGzip";

  return true;
}


CTioXMLBased::CTioXMLBased()
{
  ronly = true;

  extensions.append ("odt");
  extensions.append ("sxw");
  extensions.append ("kwd");
  extensions.append ("docx");
}


bool CTioReadOnly::save (const QString &fname)
{
  error_string = tr ("saving of this format is not supported");
  return false;
}


bool CTioXMLBased::load (const QString &fname)
{
  CZipper zipper;

  QString source_fname;
  QString ext = file_get_ext (fname);

  if (ext == "odt" || ext == "sxw" )
     source_fname = "content.xml";
  else
  if (ext == "kwd")
     source_fname = "maindoc.xml";
  else
  if (ext == "docx")
     source_fname = "word/document.xml";

  if (! zipper.read_as_utf8 (fname, source_fname))
      {
       qDebug() << "load error";
       return false;
      }

  qDebug() << "tio load::CTioXMLBased";

  CXMLWalker *walker = NULL;

  if (ext == "odt" || ext == "sxw")
     walker = new CODTXMLWalker;
  else
  if (ext == "kwd")
     walker = new CKWDXMLWalker;
  else
  if (ext == "docx")
     walker = new CDOCXXMLWalker;

  walker->doc.setContent (zipper.string_data);
  walker->step (walker->doc.documentElement());

  data = walker->data;

  delete walker;

  return true;
}


void CODTXMLWalker::step (QDomNode node)
{
  for (QDomNode n = node.firstChild(); ! n.isNull(); n = n.nextSibling())
      {
       QDomElement e = n.toElement();

       if (e.isNull())
          continue;

       if (e.nodeName() == "text:s")
          {
           QString indent = e.attribute ("text:c");

           if (! indent.isEmpty())
              {
               QString fillval;
               fillval = fillval.fill (' ', indent.toInt());
               data.append (fillval);
              }
           }
       else
       if (e.nodeName() == "text:p" ||
           e.nodeName() == "text:h")
          {
           if (! e.text().isEmpty())
              {
               data.append (e.text());
               data.append ("\n");
              }
          }

      if (e.hasChildNodes())
         step (n);
      }
}


void CKWDXMLWalker::step (QDomNode node)
{
  for (QDomNode n = node.firstChild(); ! n.isNull(); n = n.nextSibling())
      {
       QDomElement e = n.toElement();

       if (e.isNull())
          continue;

       if (e.nodeName().toLower() == "text")
          {
           data.append (e.text());
           data.append ("\n");
           }


        if (e.hasChildNodes())
           step (n);
      }
}


CTioABW::CTioABW()
{
  ronly = true;
  extensions.append ("abw");
}


bool CTioABW::load (const QString &fname)
{
  QString temp = qstring_load (fname);

  qDebug() << "tio load::CTioABW";

  CXMLWalker *walker = new CABWXMLWalker;

  walker->doc.setContent (temp);
  walker->step (walker->doc.documentElement());

  data = walker->data;

  delete walker;

  return true;
}


void CABWXMLWalker::step (QDomNode node)
{
  for (QDomNode n = node.firstChild(); ! n.isNull(); n = n.nextSibling())
      {
       QDomElement e = n.toElement();

       if (e.isNull())
          continue;

       if (e.nodeName().toLower() == "p")
          {
           data.append (e.text());
           data.append ("\n");
          }

        if (e.hasChildNodes())
           step (n);
      }
}


void CDOCXXMLWalker::step (QDomNode node)
{
  for (QDomNode n = node.firstChild(); ! n.isNull(); n = n.nextSibling())
      {
       QDomElement e = n.toElement();

       if (e.isNull())
          continue;

       if (e.nodeName().toLower() == "w:t")
          {
           data.append (e.text());
           data.append ("\n");
           }


        if (e.hasChildNodes())
           step (n);
      }
}


//www.codeguru.com/forum/archive/index.php/t-201658.html
//rewritten by Peter Semiletov
QString rtf_strip (const QString &rtf)
{
  int length = rtf.size();

  if (length < 4)
     return QString();

  int start = 0;

  start = rtf.indexOf ("\\pard");
  if ( start < 1)
     return QString();

  QString strCopy;
  strCopy.reserve (length);

  int k = 0;

  bool slash = false; //is backslash followed by the space
  bool figure_opened = false; //is opening figure brace followed by the space
  bool figure_closed = false; //is closing brace followed by the space
  bool first_space = false; //else spaces are in plain text and must be included to the result

  QChar ch;
  for (int j = start; j < length; j++)
      {

       ch = rtf.at (j);

       if (ch == '\\')//we are looking at the backslash
          {
           first_space = true;
           slash = true;
          }
       else
       if (ch == '{')
          {
           first_space = true;
           figure_opened = true;
          }
      else
      if (ch == '}')
         {
          first_space = true;
          figure_closed = true;
         }
      else
      if (ch == ' ')// &&
          //(rtf.indexOf ("\\datafield", j - 10) + 10) != j)
         {
          slash = false;
          figure_opened = false;
          figure_closed = false;
         }

      if ( ch == '\\')
         {
          QChar chr = rtf.at (j + 1);

          if (chr == '{') //if the text contains symbol '{'
           {
            slash = false;
            figure_opened = false;
            figure_closed = false;
            first_space = false;
            strCopy += '{';
            j++;
            k++;
            continue;
           }


      if (chr == '}') //if the text contains symbol '}'
         {
          slash = false;
          figure_opened = false;
          figure_closed = false;
          first_space = false;
          strCopy += '}';
          j++;
          k++;
          continue;
         }

      if (chr == '\\')//if the text contains symbol '\'
         {
          slash = false;
          figure_opened = false;
          figure_closed = false;
          first_space = false;
          strCopy += '\\';
          j++;
          continue;
         }
    }

    if (rtf.at (j) == '\\' &&
        rtf.at (j + 1) == 'p' &&
        rtf.at (j + 2) == 'a' &&
        rtf.at (j + 3) == 'r' &&
        rtf.at (j + 4) != 'd')
       {
        slash = false;
        figure_opened = false;
        figure_closed = false;
        first_space = false;
        strCopy += '\n';
        j += 4;
        continue;
       }

    if (slash == false &&
        figure_opened == false &&
        figure_closed == false &&
        ch != '\n')
        {
         if (! first_space)
            strCopy += ch;
         else
            first_space = false;
        }
  }

  return strCopy;
}


CTioRTF::CTioRTF()
{
  ronly = true;
  extensions.append ("rtf");
}


bool CTioRTF::load (const QString &fname)
{
  qDebug() << "tio load::CTioRTF";

  QByteArray ba = file_load (fname);

  QString text;
  text.reserve (ba.size());

  int i = 0;
  int l = ba.size();

  QString ansicgp;
  int n = ba.indexOf ("ansicpg");
  if (n != -1)
     {
      int m = ba.indexOf ('\\', n);
      n += 7;
      ansicgp = ba.mid (n, m - n);
      qDebug() << ansicgp;
     }

  if (ansicgp.isEmpty()) //assuming unicode
     {
      while (i < l)
             if ((ba.at(i) == '\\') && (ba.at(i + 1) == 'u'))
                {
                 QByteArray ta = ba.mid (i, 7);
                 ta = ta.mid (2, 4);
                 QChar c (ta.toInt());
                 text.append (c);
                 i += 7 + 3;
                }
             else
                 {
                  text.append (ba.at(i));
                  i++;
                 }
     }
  else
      {
       ansicgp.prepend ("CP");
       QTextCodec *codec = QTextCodec::codecForName (ansicgp.toUtf8());
       qDebug() << "not unicode!";

       while (i < l)
       if ((ba.at(i) == '\\') && (ba.at(i + 1) == '\''))
          {
           QByteArray ta = ba.mid (i, 4);
           ta = ta.mid (2, 2);
           QByteArray bh = ta.fromHex (ta);
           text.append (codec->toUnicode (bh));
           i += 4;
          }
       else
           {
            text.append (ba.at(i));
            i++;
           }
  }

  data = rtf_strip (text);

  return true;
}


void CSLAXMLWalker::step (QDomNode node)
{
  for (QDomNode n = node.firstChild(); ! n.isNull(); n = n.nextSibling())
      {
       QDomElement e = n.toElement();

       if (e.isNull())
          continue;

       if (e.nodeName().toLower() == "itext")
          {
           QString txt = e.attribute ("CH");

           if (! txt.isEmpty())
              {
               txt = txt.replace (QChar (0x5), QChar('\n'));
               data.append (txt);
               //data.append ("\n");
              }
          }

        if (e.hasChildNodes())
           step (n);
      }

  //data = data.replace ("&#x5;", "\n");

}


CTioSLA::CTioSLA()
{
  ronly = true;
  extensions.append ("sla");
}


bool CTioSLA::load (const QString &fname)
{
  QString temp = qstring_load (fname);

  qDebug() << "tio load::CTioSLA";

  CXMLWalker *walker = new CSLAXMLWalker;

  walker->doc.setContent (temp);
  walker->step (walker->doc.documentElement());

  data = walker->data;

  delete walker;

  return true;
}
